/******************************************************************************
 *  Datei Parser.C
 *  Enthlt die Funktion Parser() fr GalAsm, den Logic-Compiler / -Assembler
 *  zur Erstellung einer JEDEC-Datei fr die GAL-Programmierung.
 *
 *  Version               1.2
 *  Erstellt am           10.08.1990
 *  Letzte nderung am    3.1.1991
 *  Verwendeter Compiler: Lattice C for Atari ST, Ver 3.04
 *
 *  (c) Volker Bosch, Kaiserstrae 49, 7000 Stuttgart 70, (0711)454859
 *
 *  Bentigte Headerdateien:  GalAsm.H     Konstanten und Def. der Var.'typen
 *                            GalGlVar.H   Deklaration der globalen Variablen
 *                            HilfFktn.H   Deklaration der Hilfsfunktionen
 *
 *  Bentigte C-Module:       GalGlVar.C
 *                            HilfFktn.C
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "galasm.h"
#include "galglvar.h"
#include "hilffktn.h"


/*-------------------------------------------------------------------------
 | Funktion Parser
 | Analysiert die booleschen Gleichungen fr die Ausgnge.
 |  1. Feststellen, ob Tristate-Ausgang
 |  2. Falls 1., dann Freigabebedingung analysieren und entspr. Flags setzen:
 |        a. immer niederohmig IF(Vcc)  bzw. IF(/Gnd)
 |        b. immer hochohmig   IF(/Vcc) bzw. IF(Gnd)
 |        c. Freigabe durch /OE (Signal an Pin 11 bzw. 13, low-aktiv)
 |        d. Freigabe durch Produkt
 |  3. Falls 2.d. Analyse des Produkts:
 |        Zustand ermitteln (normal oder invertiert).
 |        Signalnamen ermitteln.
 |        Pinnummer des Signals ermitteln.
 |        Falls Signal /OE ist (Pin 11 oder 13) Flag OEinMatrix setzen.
 |        Falls Signal Clk ist (Pin 1) Flag ClkinMatrix setzen.
 |        Falls Signal ein Ausgang ist, fr diesen Ausgang das Rckkoppelflag
 |              setzen.
 |        Pinnummer mit Vorzeichen in das Feld eintragen.
 |        Fortsetzen, bis Produkt abgearbeitet.
 |  4. Falls nicht 3., dann Flags Tristate und Niederohmig setzen
 |  5. Name des Ausgangs ermitteln.
 |  6. Lfd. Nummer des Ausgangs anhand der Liste der Ausgnge ermitteln
 |  7. Analyse der Zuordnung: Normal (=) oder Registered (:=) und ggf. Flag
 |        fr Registered Output setzen.
 |  8. Analyse des Produktes, wie unter 3.
 |  9. 8. wiederholen, bis Produktterm abgearbeitet.
 | 10. Nchsten Ausgang bearbeiten, wie unter 1. bis 9.
 | 11. 10. wiederholen, bis alle Ausgnge abgearbeitet sind.
 |
 |  Rckgabewerte: Fehlercode:
 |*/
#define PARSER_ERR        20000L  /* Basis-Fehlercode fr Parser() */
                        /*  100L  Lesefehler,    in Galasm.H definiert */
                        /*  200L  Schreibfehler, in Galasm.H definiert */
#define ZU_VIEL_AUSG        300L  /* Zu viele Ausgnge angegeben */
#define FREIG_TERM          400L  /* Freigabeproduktterm falsch */
#define EING_UNBEK          500L  /* Signalname unbekannt */
#define NICHT_AUSG          600L  /* Signal ist kein Ausgang */
#define CHKFEHLER           700L  /* Kontrolle falsch */
#define ZU_VIEL_PROD        800L  /* mehr Produkte als Matrixspalten */
#define SIG_UNBEK           900L  /* Signalname nicht gefunden */
#define KEIN_FREIG         1000L  /* Freigabebedingung fehlt */
#define AUSG_N_GEFUNDEN    1100L  /* Ausgangssignal nicht gefunden */
#define ZUORDNG_FALSCH     1200L  /* Zuordnug falsch */
#define ZU_VIEL_TERME      1300L  /* mehr Produkte als Matrixspalten */
 /*
 *--------------------------------------------------------------------------*/
long Parser()
{

  char *Ptr;            /* Hilfszeiger */
  char SigName[SIGNIF_STELLEN+1];      /* Speicherfeld fr
                                          aktuell bearbeiteten Namen */
  char AusgName[SIGNIF_STELLEN+1];     /* Speicherfeld fr
                                          aktuellen Ausgangsig'namen */
  char TSFreigTerm[MAX_ANZ_SPALTEN];   /* Zwischenspeicher fr TS-Freigabe */
  char AusgNeg;         /* Flag: Ausgang invertieren */
  char SigNeg;          /* Flag: Singal invertieren */
  char OEFreig;         /* Singal ist TS, Freig. durch /OE */
  char ProdFreig;       /* Signal ist TS, Freig. d. Produkt */
  char Hochohmig;       /* Signal ist TS, immer hochohmig */
  char Niederohmig;     /* Signal ist TS, immer niederohmig */
  int  i;
  int PinNummer;
  int AusgZaehler;      /* Zhlt die Anzahl der bearbeiteten Ausgnge */
  int AusgNr;           /* Nummer des Ausgangs in Liste Ausgaenge[],  der
                           momentan bearbeitet wird   */
  int ProdTermNr;       /* Nummer der Produkttermes, der mom. bearbeitet wird*/

#ifdef GALASM_DEBUG
  printf("Bearbeiten der Produktterme...\n");
#endif
  if(! (Ptr = LeseZeile()))    /* erste Zeile lesen */
    return(PARSER_ERR + LESEFEHLER);

  AusgZaehler = 0;
  /*++++++++++++++++ Schleife ber alle Ausgnge +++++++++++++++++++++++++*/
  do {
    if (AusgZaehler == AUSGAENGE)  /* soviele Ausgnge ham'mer nich ! -> */
      return(PARSER_ERR + ZU_VIEL_AUSG);       /* Fehler ! */
    OEFreig    = 0;
    ProdFreig  = 0;
    Hochohmig  = 0;
    Niederohmig= 0;
    for(i=0;i<MAX_ANZ_SPALTEN;i++) TSFreigTerm[i]=0; /* Speicherfeld lschen*/

    Ptr = SignalNeg(Ptr,&AusgNeg);    /* Not-Operator feststellen */
    Ptr = GetSigName(Ptr,AusgName);   /* ersten Namen herauslsen */
    Ptr = LeerzUeberl(Ptr);           /* Leerzeichen berlesen */
    if(! Ptr)                         /* hoppla, Dateiende erreicht -> Fehler*/
      return(PARSER_ERR + LESEFEHLER);
    if(! stricmp("IF",AusgName)) {    /* TS-Ausgang ? */
      if(*Ptr == '(') {               /* folgt TS-Freigabebedingung ? */
        Ptr++;                        /* wenn ja, Zeiger auf nchstes Zeichen*/
        Ptr = LeerzUeberl(Ptr);       /* Leerzeichen berlesen */
        if(! Ptr)                     /* hoppla, Dateiende erreicht -> Fehler*/
          return(PARSER_ERR + LESEFEHLER);
        Ptr = SignalNeg(Ptr,&SigNeg); /* Not-Operator vorangestellt ? */
        Ptr = GetSigName(Ptr,SigName);/* Signalname herauslsen */
        Ptr = LeerzUeberl(Ptr);       /* Leerzeichen berlesen */
        if(! Ptr)                     /* hoppla, Dateiende erreicht -> Fehler*/
           return(PARSER_ERR + LESEFEHLER);
        if(! stricmp(SigName,GND_NAME)) { /* Name == GND ? */
          (SigNeg) ? (Niederohmig = 1) : (Hochohmig = 1); /* ja, Flag setzen */
          if(*Ptr++ !=')') return(PARSER_ERR + FREIG_TERM);
          Ptr = LeerzUeberl(Ptr);        /* Leerzeichen berlesen */
          if(! Ptr)                      /* Dateiende erreicht -> Fehler*/
            return(PARSER_ERR + LESEFEHLER);
        }
        else {  /* 1. else */         /* nein, dann ... */
          if(! stricmp(SigName,VCC_NAME)) {/* vielleicht Name == Vcc */
            SigNeg ? (Hochohmig = 1) : (Niederohmig = 1); /* ja, Flag setzen*/
            if(*Ptr++ !=')') return(PARSER_ERR + FREIG_TERM);
            Ptr = LeerzUeberl(Ptr);        /* Leerzeichen berlesen */
            if(! Ptr)                      /* Dateiende erreicht -> Fehler*/
              return(PARSER_ERR + LESEFEHLER);
          }
          else {  /* 2. else */       /* nein, auch nicht, ... */
            i = 0;   /* dann den Namen des Signals am OE Pin suchen */
            while((Eingaenge[i].PinNummer != OEPin) && (i<EINGAENGE)) {
              i++;
            }
            if(i==EINGAENGE)       /*Pin nicht gefunden -> Fehler ! */
              return(PARSER_ERR + EING_UNBEK);
            if((! stricmp(SigName,Eingaenge[i].SigName)) && (SigNeg)) {
                                      /* Freigabe durch /OE ? */
              Eingaenge[i].Genutzt = 1;  /* Flag setzten */
              if(*Ptr == ')' )  {     /* dann folgt kein Produktterm ! */
                Ptr++;                /* wenn ja, Zeiger auf nchstes Zeichen*/
                Ptr = LeerzUeberl(Ptr);/* Leerzeichen berlesen */
                if(! Ptr)              /* hoppla, Dateiende erreicht */
                  return(PARSER_ERR + LESEFEHLER);  /*  -> Fehler*/
                OEFreig = 1;           /* -> OEFreigabeflag setzen */
              }
            }
            if( ! OEFreig) {    /* nur Produkttermfreigabe mglich */
              ProdFreig  = 1;   /* -> entspr. Flag setzen */
              PinNummer = GetEingPinNr(SigName); /* Pinnummer feststellen */
              if(! PinNummer) {                  /* Signal ist kein Eingang */
                PinNummer = GetAusgPinNr(SigName); /* Pinnr. feststellen */
                if(! PinNummer)                  /* Sig. ist kein Ausgang */
                  return(PARSER_ERR + SIG_UNBEK);             /* -> Fehler */
              }
              if((PinNummer == GndPin) ||        /* Kann und darf nicht */
                 (PinNummer == VccPin)    )      /* sein ! -> */
                    return(PARSER_ERR + CHKFEHLER); /* Fehler ! */
              TSFreigTerm[0] =         /* sonst, Pinnummer und akt. Pegel */
                (SigNeg) ? (-PinNummer) : (PinNummer);   /* in Feld eintr. */
              i = 1;                /* Laufvariable initialisieren */
              /*-------------- den restlichen Produktterm abarbeiten -----*/
              while(*Ptr == '*') {
                if(i == AnzSpalten)    /* mehr Produkkterme als Spalten -> */
                  return(PARSER_ERR + ZU_VIEL_PROD);          /* Fehler */
                Ptr++;              /* Zeiger auf nchstes Zeichen setzen */
                Ptr = LeerzUeberl(Ptr); /* und fhrende Leerzeichen berlesen*/
                if(! Ptr)               /* Dateiende -> Fehler */
                  return(PARSER_ERR + LESEFEHLER);
                Ptr = SignalNeg(Ptr,&SigNeg);  /* Not-Operator ? */
                Ptr = GetSigName(Ptr,SigName); /* Signalname ermitteln */
                Ptr = LeerzUeberl(Ptr);        /* Leerzeichen berlesen */
                if(! Ptr)                      /* Dateiende -> Fehler */
                  return(PARSER_ERR + LESEFEHLER);
                PinNummer = GetEingPinNr(SigName);/* Pinnummer feststellen */
                if(! PinNummer) {                 /* Signal ist kein Eingang */
                  PinNummer = GetAusgPinNr(SigName); /* Pinnr. feststellen */
                  if(! PinNummer)           /* Sig. ist auch kein Ausgang -> */
                    return(PARSER_ERR + SIG_UNBEK);   /* Fehler */
                }
                if((PinNummer == GndPin) ||   /* Kann und darf nicht */
                   (PinNummer == VccPin)    ) /* sein -> */
                      return(PARSER_ERR + CHKFEHLER);   /* Fehler ! */
                TSFreigTerm[i++] =            /* sonst, akt. Logikpegel und */
                  (SigNeg) ? (-PinNummer) : (PinNummer); /* Pinnr. speichern */
              }  /* while */
              /*------ Produktterm abgearbeitet -----*/
              if(*Ptr++ != ')')        /* Freigabeproduktterm nicht Korrekt */
                return(PARSER_ERR + FREIG_TERM);  /* abgeschlossen -> Fehler */
              Ptr = LeerzUeberl(Ptr);  /* Leerzeichen berlesen */
              if(! Ptr)                /* hoppla, Dateiende erreicht */
                return(PARSER_ERR + LESEFEHLER);    /*  -> Fehler*/
            }  /* if (! OEFreig) */
          } /* 2. else */
        } /* 1. else */
      } /* if(*Ptr == '(') */
      else                   /* Fehler, da Freigabebedingung fehlt */
         return(PARSER_ERR + KEIN_FREIG);
      Ptr = SignalNeg(Ptr,&AusgNeg);  /* Not-Operator vor Ausgangsterm ? */
      Ptr = GetSigName(Ptr,AusgName); /* Name des Ausgagsterms */
    }
    AusgNr = -1;                    /* Zur Fehlererkennung */
    /*-- Postion des Ausgangs in der Liste der Ausgnge suchen ---*/
    for(i=0;((i<AUSGAENGE) && (Ausgaenge[i].PinNummer));i++) {
      if(! stricmp(AusgName,Ausgaenge[i].SigName)) {
        AusgNr = i;
      }
    }
    if(AusgNr < 0)        /*Fehler, Ausgangssig. nicht gefunden*/
       return(PARSER_ERR + AUSG_N_GEFUNDEN);
    if(AusgNeg) {    /* Ausgang ggf. invertieren */
      Ausgaenge[AusgNr].Negieren =  !(Ausgaenge[AusgNr].Negieren);
    }
    Ausgaenge[AusgNr].Genutzt = 1;       /* Ausgang wird genutzt */
    Ausgaenge[AusgNr].Ausg = 1;          /* Ausgang wird als solcher genutzt */
    Ausgaenge[AusgNr].OEFreig = OEFreig;          /* - " -  -  "  -   */
    Ausgaenge[AusgNr].ProdFreig = ProdFreig;      /* - " -  -  "  -   */
    Ausgaenge[AusgNr].Hochohmig = Hochohmig;      /* - " -  -  "  -   */
    Ausgaenge[AusgNr].Niederohmig = Niederohmig;  /* - " -  -  "  -   */
    ProdTermNr = 0;
    if(ProdFreig) {
      for(i=0;i<AnzSpalten;i++) {    /* TS-Freigabe-Produktterm kopieren */
        LogikMatrix[AusgNr][ProdTermNr][i] = TSFreigTerm[i];
      }
      ProdTermNr = 1;
    }
    Ptr = LeerzUeberl(Ptr);    /* Zeiger auf nchstes druckendes Zeichen */
    if(! Ptr) return(PARSER_ERR + LESEFEHLER);  /* oh, ein Fehler ! */
    switch (*Ptr) {            /* Zuordnugsart feststellen */
      case ':' : {
        if(*(Ptr+1) == '=') {
          Ausgaenge[AusgNr].Reg = 1; /* Register Ausgang */
          for(i=0;i<EINGAENGE;i++) {
            if(Eingaenge[i].PinNummer == ClkPin) Eingaenge[i].Genutzt = 1;
          }
          for(i=0;i<EINGAENGE;i++) {
            if(Eingaenge[i].PinNummer == OEPin) Eingaenge[i].Genutzt = 1;
          }
        }
        else return(PARSER_ERR + ZUORDNG_FALSCH);
        Ptr+=2;               /* nchstes Zeichen */
        break;
      }
      case '=' : {            /* normaler Ausgang */
        Ausgaenge[AusgNr].Ausg = 1;      /* Flag setzen */
        Ptr++;                           /* nchstes Zeichen */
        break;
      }
      default : return(PARSER_ERR + ZUORDNG_FALSCH);       /* Ein Fehler ! */
    }
    Ptr = LeerzUeberl(Ptr);        /* auf nchstes druckendes Zeichen */
    if(! Ptr)                      /* hoppla, unerwartetes Dateiende */
      return(PARSER_ERR + LESEFEHLER);
    /*============ Schleife ber alle ODER-Verknpfungen ============*/
    do {
  #ifdef GALASM_DEBUG
      printf("%d. Produkterm von %d. Ausgang \r",ProdTermNr+1,AusgZaehler+1);
  #endif
      if (ProdTermNr == ANZ_PROD) return(PARSER_ERR + ZU_VIEL_TERME);
      if(*Ptr == '+') Ptr++;      /* ODER-Verknpfung berlesen */
      i = 0;  /* Laufvariable fr die einzelenen AND zu verknpfenden Terme */
      /*-------- Schleife ber alle AND zu verknpfenden Terme ---------*/
      do {
        if(i == AnzSpalten)         /* mehr Produkkterme als Spalten */
          return(PARSER_ERR + ZU_VIEL_PROD);
        if (*Ptr == '*') Ptr++;        /* '*' berlesen, falls vorhanden */
        Ptr = LeerzUeberl(Ptr);        /* Zeiger auf nchstes Zeichen */
        if(! Ptr)                      /* Dateiende erreicht -> Fehler ! */
          return(PARSER_ERR + LESEFEHLER);
        Ptr = SignalNeg(Ptr,&SigNeg);  /* Signal invers ? */
        Ptr = GetSigName(Ptr,SigName); /* Signalname feststellen */
        Ptr = LeerzUeberl(Ptr);        /* Leerzeichen berlesen */
        PinNummer = GetEingPinNr(SigName); /* Pinnummer feststellen */
        if(! PinNummer) {                  /* Signal ist kein Eingang */
          PinNummer = GetAusgPinNr(SigName); /* Pinnr. feststellen */
          if(! PinNummer)      /* Sig. ist kein Ausgang -> Fehler */
            return(PARSER_ERR + SIG_UNBEK);
        }
        if((PinNummer == GndPin) ||      /* Das kann ja wohl */
           (PinNummer == VccPin)     )   /* nicht sein ! -> */
          return(PARSER_ERR + CHKFEHLER);             /* Fehler ! */
        if(SigNeg) PinNummer = -PinNummer;
        LogikMatrix[AusgNr][ProdTermNr][i++] =  /* Pinnr. und Logikpegel */
          PinNummer;                            /* eintragen */
      } while(Ptr && (*Ptr == '*'));        /* solange Dateiende noch nicht */
                                    /* und AND-Verknpfungen gefunden werden */
      /*-------- Schleife ber alle AND zu verknpfenden Terme ---------*/
      ProdTermNr++;   /* nchste Summe von Produkten */
    } while (Ptr && (*Ptr == '+'));          /* solange Dateiende noch nicht */
                                             /* erreicht und '+' vorhanden */
    /*============ Schleife ber alle ODER-Verknpfungen ============*/
  AusgZaehler++;
  Ptr = LeerzUeberl(Ptr);        /* Zeiger auf nchstes Zeichen */
  } while(Ptr);                  /* wiederholen, bis Dateiende erreicht */
  /*++++++++++++++++ Schleife ber alle Ausgnge +++++++++++++++++++++++++*/
  gotoxy(0,MSG_ZEILE);
  end_screen();
  return(KEIN_FEHLER);
}


